﻿/**
 * @file canethlib.cs
 *
 * CAN .NET API for proconX CAN-ETH gateway
 *
 * @if NOTICE
 *
 * Copyright (c) 2021 proconX Pty Ltd. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PROCONX AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL PROCONX OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * For additional information see http://www.proconx.com
 *
 * @endif
 */

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;


namespace CanEth
{
    /// <summary>
    /// Class representing a single CAN message
    /// </summary>
    public class CanEthMsg
    {
        int id;      
        int dlc;        
        Byte[] data = new Byte[8]; // Message byte buffer

        public bool Ext;       ///< Boolean flag to indicate extended frame, false = 11-bit identifier, true = 29-bit identifier
        public bool Rtr;       ///< Boolean flag to indicate remote transmition bit, false = standard frame, true = remote tranmission request


        /// <summary> CAN ID property </summary>
        /// <value> Max ID is 0x7FF if ext=false or 0x1FFFFFF if ext=true </value>        
        public int Id
        {
            get { return this.id; }
            set
            {
                if (value > 0x1FFFFFF)
                    throw new ArgumentOutOfRangeException("Max extended id is 0x1FFFFFF");
                this.id = value;
            }
        }


        /// <summary> DLC property </summary>
        /// <value> Number of data bytes (0-8) </value>        
        public int Dlc
        {
            get { return this.dlc; }
            set
            {
                if (value < 0 || value > data.Length)
                    throw new ArgumentOutOfRangeException("DLC must be within 0 to 8");
                this.dlc = value;
            }
        }

        /// <summary> Message buffer accessor </summary>
        public byte this[int i]
        {
            get 
            {
                if (i < 0 || i >= data.Length)
                    throw new IndexOutOfRangeException("Index must be within 0 to 8");
                return data[i]; 
            }
            set 
            {
                if (i < 0 || i >= data.Length)
                    throw new IndexOutOfRangeException("Index must be within 0 to 8");
                data[i] = value; 
            }
        }


        /// <summary> Creates a CAN message </summary>
        public CanEthMsg(int id = 0, Byte[] data = null, bool ext = false, bool rtr = false)
        {
            this.Id = id;
            if (data == null)
            {
                this.Dlc = 0;
            }
            else
            {
                this.Dlc = data.Length;
                data.CopyTo(this.data, 0);
            }
            this.Ext = ext;
            this.Rtr = rtr;
        }


        public override string ToString()
        {
            string s = String.Format(Ext ? "{0:X7}{1}:" : "{0:X3}{1}:", Id, Rtr ? "R" : "");
            foreach (var val in data)
                s += String.Format(" {0:X2}", val);
            return s;
        }


        internal List<byte> asList()
        {
            var msg = new List<byte>();

            if (!Ext && id > 0x7FF)
                throw new ArgumentOutOfRangeException("Max standard CAN ID is 0x7FF");
            msg.AddRange(BitConverter.GetBytes(id)); // CAN ID in little-endian format, max ID is 0x7FF if ext=0 or 0x1FFFFFF if ext=1
            msg.Add((byte)dlc); // Number of data bytes (0-8)
            for (int i = 0; i < data.Length; i++)
            {
                if (i < dlc)
                    msg.Add((byte)data[i]); // Message byte buffer
                else
                    msg.Add(0); // Prevent data leakage
            }
            msg.Add(Ext ? (byte)1 : (byte)0); // Boolean flag to indicate extended frame, 0 = 11-bit identifier, 1 = 29-bit identifier
            msg.Add(Rtr ? (byte)1 : (byte)0); // Boolean flag to indicate remote transmition bit, 0 = standard frame, 1 = remote tranmission request

            return msg;
        }
    }


    /// <summary>
    /// Class representing a CAN-ETH variable-length UDP message.
    /// It encapsulates multiple CAN messages and can be accessed similar to a List container.
    /// </summary>
    public class CanEthFrame: IEnumerable
    {
        List<CanEthMsg> messages = new List<CanEthMsg>();

        public const int UDP_PORT = 11898;
        public const int MAX_MSG = 96;  // v1 has up to 16 messages, v2 has up to 96
        public const string MAGIC_ISO11898_ID = "ISO11898";  // Must be char sequence 'ISO11898'


        /// <summary> Creates an empty frame </summary>
        public CanEthFrame()
        {
            Clear();
        }


        /// <summary> Creates a frame from a UDP message's byte array </summary>
        public CanEthFrame(byte[] datagram)
        {
            if (validate(datagram))
            {
                int count = datagram[9];
                int ptr = 10;

                for (int i = 0; i < count; i++)
                {
                    var msg = new CanEthMsg();

                    msg.Id = BitConverter.ToInt32(datagram, ptr);
                    ptr += 4;
                    msg.Dlc = datagram[ptr++];
                    for (int j = 0; j < 8; j++)
                        msg[j] = datagram[ptr++];
                    msg.Ext = datagram[ptr++] != 0;
                    msg.Rtr = datagram[ptr++] != 0;
                    messages.Add(msg);
                }
            }
            else
                throw new ApplicationException("Invalid CAN-ETH datagram");
        }


        /// <summary>
        /// Returns an enumerator that iterates through the CAN message in this frame.
        /// </summary>
        /// <returns>A System.Collections.Generic.List.Enumerator</returns>
        public IEnumerator GetEnumerator()
        {
            return messages.GetEnumerator();
        }


        /// <summary> Message container accessor </summary>
        public CanEthMsg this[int index]
        {
            get
            {
                return messages[index];
            }
        }


        /// <summary>
        /// Gets the number of messages contained in this frame.
        /// </summary>
        /// <returns>The number of elements contained in this frame.</returns>
        public int Count
        {
            get
            {
                return messages.Count;
            }
        }


        /// <summary>
        /// Adds a CAN message to UDP frame
        /// </summary>
        /// <param name="msg">CAN message</param>
        public void Add(CanEthMsg msg)
        {
            if (messages.Count > 96)
                throw new IndexOutOfRangeException("Max frame couunt is 96");
            messages.Add(msg);
        }


        /// <summary>
        /// Removes all CAN messages from UDP frame
        /// </summary>
        public void Clear()
        {
            messages.Clear();
        }


        /// <summary>
        /// Validates a received CAN-ETH UDP frame
        /// </summary>
        /// <param name="datagram">CAN-ETH UDP frame</param>
        /// <returns>True if frame validates OK</returns>
        private bool validate(byte[] datagram)
        {
            // Check minimum length
            if (datagram.Length <= 10)
                return false;
            // Check for magic ID
            if (MAGIC_ISO11898_ID != Encoding.ASCII.GetString(datagram, 0, MAGIC_ISO11898_ID.Length))
                return false;
            // Check for supported version
            if (datagram[8] != 1 && datagram[8] != 2)
                return false;
            // Check count within bounds
            if (datagram[9] > MAX_MSG)
                return false;
            // Check minimum size
            if (datagram.Length < 10 + datagram[9] * 15) // 10 bytes header plus 15 bytes for each message
                return false;

            // All tests passed, OK
            return true;
        }


        /// <summary>
        /// Packs all CAN frames into a byte array
        /// </summary>
        /// <returns>Array of bytes</returns>
        public byte[] GetBytes()
        {
            var datagram = new List<byte>();

            datagram.AddRange(Encoding.ASCII.GetBytes(MAGIC_ISO11898_ID));
            datagram.Add(messages.Count > 16 ? (Byte)2 : (Byte)1); // v1 has up to 16 messages, v2 has up to 96
            datagram.Add((Byte)messages.Count);
            foreach (var msg in messages)
            {
                datagram.AddRange(msg.asList());
            }
            datagram.Add(0xFF); // EOL marker

            return datagram.ToArray();
        }
    }
}
