#!/usr/bin/env python
"""
A simple CANopen Master which puts CANopen nodes into operational mode for the CAN-ETH gateway.
"""
# Copyright (c) 2023 proconX Pty Ltd. All rights reserved.
#
# The following source file constitutes example program code and is
# intended merely to illustrate useful programming techniques.  The user
# is responsible for applying the code correctly.
#
# THIS SOFTWARE IS PROVIDED BY PROCONX AND CONTRIBUTORS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL PROCONX OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys, socket
sys.path.append('.')
from canethlib import *
from enum import IntEnum


CANopen_NMT_CONTROL = 0x000
CANopen_NMT_HEARTBEAT = 0x700
CANopen_MAX_NODES = 127


class CANopenNmtStates(IntEnum):
    BOOT_UP = 0
    STOPPED = 4
    OPERATIONAL = 5
    PRE_OPERATIONAL = 0x7F
    def __str__(self):
        return self._name_


class CANopenNmtCommand(IntEnum):
    Start_Remote_Node = 0x1
    Stop_Remote_Node = 0x2
    Enter_Pre_Operational_State = 0x80
    Reset_Node = 0x81
    Reset_Communication = 0x82
    def __str__(self):
        return self._name_
    

def main():
    """A simple CANopen Master which puts CANopen nodes into operational mode for the CAN-ETH gateway."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) # Create a UDP socket
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind(('', CANETH_PORT)) # Bind this socket to UDP port 11898
    while True:
        # Receive UDP datagram
        datagram, addr = sock.recvfrom(CanEthFrame.MAX_DATAGRAM) 
        if datagram:
            for msg in CanEthFrame(datagram):
                # Check for NMT heartbeat messages
                if (msg.id > CANopen_NMT_HEARTBEAT and msg.id <= CANopen_NMT_HEARTBEAT + CANopen_MAX_NODES) and len(msg) == 1:
                    slaveNodeId = msg.id - CANopen_NMT_HEARTBEAT
                    slaveNodeState = CANopenNmtStates(msg[0])
                    print('CANopen NMT message from node ID %d: %s' % (slaveNodeId, slaveNodeState))
                    # Start pre-operational nodes
                    if slaveNodeState == CANopenNmtStates.PRE_OPERATIONAL: 
                            # Create a NMT Module Control Service message to start a node
                            masterMsg = CanMsg(CANopen_NMT_CONTROL, (slaveNodeId, int(CANopenNmtCommand.Start_Remote_Node)), ext=False)
                            masterFrame = CanEthFrame() # Create a container for multiple CAN messages
                            masterFrame.append(masterMsg) # Add one message to the container
                            sock.sendto(masterFrame.pack(), (addr[0], CANETH_PORT)) # Send UDP datagram
                            print('NMT Module Control Service message to start node %d sent to %s' % (slaveNodeId, addr[0]))

if __name__ == "__main__":
    main()

